"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.diff = diff;
exports.diffElementInfo = diffElementInfo;
exports.diffInfoOf = diffInfoOf;

var _HTypes = require("./HTypes");

var _HDocument = require("./HDocument");

var _lodash = require("lodash");

var _HVisit = require("./HVisit");

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * Returns a list of HDocOperations that if applied
 * will transform baseDoc in laterDoc. laterDoc
 * is assumed to have the same root element and schema as
 * baseDoc
 *
 * @param {INormalizedDocument<MapsInterface, U>} baseDoc
 * @param {INormalizedDocument<MapsInterface, U>} laterDoc
 * @returns {HDocOperation<MapsInterface, any, U>[]}
 */
function diff(baseDoc, laterDoc) {
  /**
   * We visit laterDoc breadth first, and for each element visited
   * we ensure that the info is up to date and that the children
   * in laterDoc are all in the correct position.
   *
   * If a move is needed info changes are also performed to avoid
   * a second change info command. At the end of the visit some
   * children arrays may have additional elements in the list.
   *
   * These elements will either move somewhere else later in the
   * tree visit, if they are still present in laterdoc, or they
   * will be deleted after the visit.
   */
  if (!(baseDoc.schema === laterDoc.schema && baseDoc.rootType === laterDoc.rootType && baseDoc.rootId === laterDoc.rootId)) {
    return [];
  }

  const mutableDoc = (0, _HDocument.mutableDocument)(baseDoc);
  (0, _HVisit.visitDocument)(laterDoc, (doc, nodeType, nodeId) => {
    const destElement = (0, _HDocument.mappedElement)(laterDoc.maps, nodeType, nodeId);
    let mutableElement = (0, _HDocument.mappedElement)(mutableDoc.maps, nodeType, nodeId);

    if (!(0, _HDocument.isParentedId)(destElement) || !(0, _HDocument.isParentedId)(mutableElement)) {
      throw new ReferenceError(`Node ${nodeType}:${nodeId} not found`);
    } // 1. If we are an existing node, check if the info fields should be
    // updated


    const nodePath = (0, _HDocument.pathForElementWithId)(mutableDoc, nodeType, nodeId);

    if ((0, _HDocument.hasMappedElement)(baseDoc.maps, nodeType, nodeId)) {
      const infoChanges = diffInfoOf(mutableDoc, laterDoc, nodeType, nodeId);

      if (Object.keys(infoChanges).length > 0) {
        const changeElementCmd = {
          __typename: _HTypes.HDocCommandType.CHANGE_ELEMENT,
          targetElement: {
            __typename: nodeType,
            _id: nodeId
          },
          path: nodePath,
          changes: _objectSpread({
            __typename: destElement.__typename
          }, infoChanges)
        };
        mutableDoc.changeElement(changeElementCmd);
      }
    } // 2. Iterate across the children


    const {
      schema
    } = doc;

    for (const linkFieldName in schema.types[nodeType]) {
      if (linkFieldName === 'parentId') continue;
      const fieldLink = schema.types[nodeType][linkFieldName];

      if (Array.isArray(fieldLink)) {
        const {
          __schemaType: childType
        } = fieldLink[0];
        const destChildrenIds = destElement[linkFieldName];

        for (let i = 0; i < destChildrenIds.length; i++) {
          const destChildId = destChildrenIds[i]; // Every iteration of dest child, there is a chance that the mutable
          // element has changed, so I need to refresh to the latest reference

          mutableElement = (0, _HDocument.mappedElement)(mutableDoc.maps, nodeType, nodeId);
          const mutableChildrenIds = mutableElement[linkFieldName];
          const mutableChildId = i < mutableChildrenIds.length ? mutableChildrenIds[i] : null;
          const destChild = (0, _HDocument.mappedElement)(laterDoc.maps, childType, destChildId);

          if (!(0, _HDocument.isParentedId)(destChild)) {
            throw new ReferenceError(`Child node ${childType}:${destChildId} not found`);
          }

          if (destChildId !== mutableChildId) {
            // No else branch needed, the position is the same and the visit to the child node
            // will take care of the potential differences int he data within the child node
            if ((0, _HDocument.hasMappedElement)(mutableDoc.maps, childType, destChildId)) {
              // Node exists in the document, move it from there
              const childInfoDiff = diffInfoOf(mutableDoc, laterDoc, childType, destChildId);
              const moveChildCmd = {
                __typename: _HTypes.HDocCommandType.MOVE_ELEMENT,
                targetElement: {
                  _id: destChildId,
                  __typename: childType
                },
                fromPath: (0, _HDocument.pathForElementWithId)(mutableDoc, childType, destChildId),
                toParentPath: (0, _HDocument.pathForElementWithId)(mutableDoc, nodeType, nodeId),
                toPosition: [linkFieldName, i],
                changes: Object.keys(childInfoDiff).length > 0 ? _objectSpread({
                  __typename: destChild.__typename
                }, (0, _lodash.omit)(childInfoDiff, '__typename')) : undefined
              };
              mutableDoc.moveElement(moveChildCmd);
            } else {
              // New element, let's add the basic info from it,
              // emptying children links
              const elementInfo = _objectSpread({}, destChild);

              for (const childLinkFieldName in schema.types[childType]) {
                if (childLinkFieldName === 'parentId') continue;
                const childFieldLink = schema.types[childType][childLinkFieldName];
                elementInfo[childLinkFieldName] = Array.isArray(childFieldLink) ? [] : null;
              }

              const addChildCmd = {
                __typename: _HTypes.HDocCommandType.INSERT_ELEMENT,
                parentPath: nodePath,
                position: [linkFieldName, i],
                element: elementInfo,
                targetElement: {
                  __typename: childType,
                  _id: destChild._id
                }
              };
              mutableDoc.insertElement(addChildCmd);
            }
          }
        }
      } else {
        if (destElement[linkFieldName] !== mutableElement[linkFieldName]) {
          const changeLinkFieldCmd = {
            __typename: _HTypes.HDocCommandType.CHANGE_ELEMENT,
            path: (0, _HDocument.pathForElementWithId)(mutableDoc, nodeType, nodeId),
            targetElement: {
              __typename: nodeType,
              _id: nodeId
            },
            changes: {
              __typename: destElement.__typename,
              [linkFieldName]: destElement[linkFieldName]
            }
          };
          mutableDoc.changeElement(changeLinkFieldCmd);
        }
      }
    }
  }, {}); // After replicating the destination tree, I can go through the mutable document
  // depth first and delete any element that was in the base tree but is not in
  // the destination tree

  (0, _HVisit.visitDocument)(mutableDoc, (doc, nodeType, nodeId) => {
    if (!(0, _HDocument.hasMappedElement)(laterDoc.maps, nodeType, nodeId)) {
      const deleteElementCmd = {
        __typename: _HTypes.HDocCommandType.DELETE_ELEMENT,
        path: (0, _HDocument.pathForElementWithId)(doc, nodeType, nodeId),
        targetElement: {
          __typename: nodeType,
          _id: nodeId
        }
      };
      mutableDoc.deleteElement(deleteElementCmd);
    }
  }, {}, _HTypes.DocumentVisitTraversal.DEPTH_FIRST);
  return mutableDoc.changes;
}
/**
 * Given a base version and a later version of a document element
 * with schema schema and type elementType, returns a Partial version
 * of the element containing the info fields that have changed between baseEl
 * and laterEl.
 *
 * @param {IDocumentSchema<MapsInterface, U>} schema
 * @param {U} elementType
 * @param {T} baseEl
 * @param {T} laterEl
 * @returns {Partial<T>}
 */


function diffElementInfo(schema, elementType, baseEl, laterEl) {
  const infoDiff = {};
  const elementLinkedFields = schema.types[elementType];

  if (!elementLinkedFields) {
    return infoDiff;
  }

  const fieldsChecked = new Set();

  for (const fieldName in baseEl) {
    if (fieldName === 'parentId' || fieldName in elementLinkedFields) {
      continue;
    }

    fieldsChecked.add(fieldName);
    const baseVal = baseEl[fieldName];
    const laterVal = laterEl[fieldName];

    if (!(0, _lodash.isEqual)(baseVal, laterVal)) {
      infoDiff[fieldName] = laterVal;
    }
  }

  for (const fieldName in laterEl) {
    if (fieldName === 'parentId' || fieldName in elementLinkedFields || fieldsChecked.has(fieldName)) {
      continue;
    } // If the field is not in the set of checked fields,
    // it was undefined in base but is defined in later


    infoDiff[fieldName] = laterEl[fieldName];
  }

  return infoDiff;
}

function diffInfoOf(baseDoc, laterDoc, elementType, elementId) {
  if (!(0, _HDocument.hasMappedElement)(baseDoc.maps, elementType, elementId) || !(0, _HDocument.hasMappedElement)(laterDoc.maps, elementType, elementId)) {
    return {};
  }

  const baseEl = (0, _HDocument.mappedElement)(baseDoc.maps, elementType, elementId);
  const laterEl = (0, _HDocument.mappedElement)(laterDoc.maps, elementType, elementId);
  return diffElementInfo(baseDoc.schema, elementType, baseEl, laterEl);
}
//# sourceMappingURL=HDiff.js.map