"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.threeWayMerge = threeWayMerge;
exports.mergeElementInfo = mergeElementInfo;

var _uuid = require("uuid");

var _nodeDiff = require("node-diff3");

var _HDocument = require("./HDocument");

var _HTypes = require("./HTypes");

var _HVisit = require("./HVisit");

var _HDiff = require("./HDiff");

let _Symbol$iterator;

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function isDataValue(obj) {
  return typeof obj === 'string' || typeof obj === 'number' || typeof obj === 'boolean' || obj instanceof Date;
}
/**
 * Merges atomic values, given a base value, and a
 * mine and their versions for a three-way merge.
 *
 * If a conflict occurs that cannot be fixed, it returns a IValueConflict
 * record. If no conflict occurs the merged value is returned instead.
 *
 * @param {T} baseValueInp
 * @param {T} myValueInp
 * @param {T} theirValueInp
 * @returns {IValueConflict<T> | T}
 */


function mergeDataValues(baseValueInp, myValueInp, theirValueInp) {
  const baseValue = baseValueInp instanceof Date ? baseValueInp.getTime() : baseValueInp;
  const myValue = myValueInp instanceof Date ? myValueInp.getTime() : myValueInp;
  const theirValue = theirValueInp instanceof Date ? theirValueInp.getTime() : theirValueInp;

  if (theirValue === myValue) {
    return myValueInp;
  }

  if (baseValue === myValue) {
    return theirValueInp;
  }

  if (baseValue === theirValue) {
    return myValueInp;
  }

  let mergedValue;

  if (typeof baseValue === 'boolean' && typeof theirValue === 'boolean' && typeof myValue === 'boolean') {
    mergedValue = myValue;
  } else if (typeof baseValue === 'number' && typeof theirValue === 'number' && typeof myValue === 'number') {
    const myDiff = Math.abs(theirValue - baseValue);
    const theirDiff = Math.abs(theirValue - baseValue);

    if (myDiff > theirDiff) {
      mergedValue = myValueInp;
    } else if (myDiff < theirDiff) {
      mergedValue = theirValueInp;
    } else {
      mergedValue = myValue < theirValue ? myValueInp : theirValueInp;
    }
  } else if (typeof baseValue === 'string' && typeof theirValue === 'string' && typeof myValue === 'string') {
    const res = (0, _nodeDiff.diff3Merge)(myValue, baseValue, theirValue, {
      stringSeparator: ''
    });
    mergedValue = res.every(region => typeof region === 'string') ? res.join('') : myValue < theirValue ? myValueInp : theirValueInp;
  } else {
    mergedValue = myValue === theirValue && myValue === baseValue ? baseValueInp : myValue < theirValue ? myValueInp : theirValueInp;
  }

  return {
    baseValue: baseValueInp,
    conflictValues: [myValueInp, theirValueInp],
    mergeStatus: _HTypes.MergeStatus.open,
    mergedValue
  };
}

const defaultHooks = {
  onIncompatibleElementVersions: (elementType, elementId, parentPath, position, versionMoved, mergeCtx) => {
    const {
      newElementId
    } = reIdElementSubtree(mergeCtx, versionMoved === _HTypes.ProcessingOrderFrom.left ? _HTypes.ProcessingOrderFrom.right : _HTypes.ProcessingOrderFrom.left, elementType, elementId);
    const cloneId = newElementId;
    const elementConflicts = mergeCtx.conflicts[elementType].get(elementId) || {};
    elementConflicts.positionConflicts = {
      clonedElements: [cloneId],
      mergeStatus: _HTypes.MergeStatus.autoMerged
    };
    mergeCtx.conflicts[elementType].set(elementId, elementConflicts);
    let currentId = null;
    let currentType = null;

    try {
      const {
        _id,
        __typename
      } = (0, _HDocument.idAndTypeForPath)(mergeCtx.mergedDoc, parentPath.concat(position));
      currentId = _id;
      currentType = __typename;
    } catch (err) {// it's possible there is no node at the path, in which case
      // let's swallow the integrity exception
    }

    if (currentType !== elementType || currentId !== elementId) {
      const moveElementCmd = {
        __typename: _HTypes.HDocCommandType.MOVE_ELEMENT,
        fromPath: (0, _HDocument.pathForElementWithId)(mergeCtx.mergedDoc, elementType, elementId),
        toParentPath: parentPath,
        toPosition: position,
        changes: {
          __typename: elementType
        },
        targetElement: {
          __typename: elementType,
          _id: elementId
        }
      };
      mergeCtx.mergedDoc.moveElement(moveElementCmd);
    }

    if (versionMoved === _HTypes.ProcessingOrderFrom.left) {
      mergeCtx.myElementsMergeState.set(getElementUid(elementType, cloneId), mergeCtx.myElementsMergeState.get(getElementUid(elementType, elementId)));
    } else {
      mergeCtx.theirElementsMergeState.set(getElementUid(elementType, cloneId), mergeCtx.theirElementsMergeState.get(getElementUid(elementType, elementId)));
    }

    return {
      advancedMergingIndex: true,
      rebasedIds: [{
        rebasedSide: versionMoved === _HTypes.ProcessingOrderFrom.left ? _HTypes.ProcessingOrderFrom.right : _HTypes.ProcessingOrderFrom.left,
        _id: elementId,
        newId: cloneId
      }]
    };
  },
  moveToMergePosition: (elementType, elementId, toParentPath, toPosition, mergeContext) => {
    const moveCmd = {
      __typename: _HTypes.HDocCommandType.MOVE_ELEMENT,
      fromPath: (0, _HDocument.pathForElementWithId)(mergeContext.mergedDoc, elementType, elementId),
      toParentPath,
      toPosition,
      targetElement: {
        __typename: elementType,
        _id: elementId
      }
    };
    mergeContext.mergedDoc.moveElement(moveCmd);
  },
  mergeElementInfo: (mergeContext, elementType, base, a, b, ignoreFields) => {
    let elementInfoDiff = null;

    if (a && b && base) {
      const mergeElement = (0, _HDocument.mappedElement)(mergeContext.mergedDoc.maps, elementType, base._id);
      const mergeRes = mergeElementInfo(mergeContext.mergedDoc.schema, elementType, base, a, b, ignoreFields);

      if (mergeRes.conflicts) {
        const elementConflicts = mergeContext.conflicts[elementType].get(base._id) || {};
        mergeContext.conflicts[elementType].set(base._id, _objectSpread(_objectSpread({}, elementConflicts), {}, {
          infoConflicts: mergeRes.conflicts
        }));
      }

      elementInfoDiff = (0, _HDiff.diffElementInfo)(mergeContext.mergedDoc.schema, elementType, mergeElement, mergeRes.mergedElement);
    } else if (base && (a || b)) {
      const mergeElement = (0, _HDocument.mappedElement)(mergeContext.mergedDoc.maps, elementType, base._id);
      const laterEl = a ? a : b;
      elementInfoDiff = (0, _HDiff.diffElementInfo)(mergeContext.mergedDoc.schema, elementType, mergeElement, laterEl);
    }

    if (elementInfoDiff && Object.keys(elementInfoDiff).length > 0) {
      const changeCmd = {
        __typename: _HTypes.HDocCommandType.CHANGE_ELEMENT,
        path: (0, _HDocument.pathForElementWithId)(mergeContext.mergedDoc, elementType, base._id),
        changes: _objectSpread(_objectSpread({}, elementInfoDiff), {}, {
          __typename: elementType
        }),
        targetElement: {
          __typename: elementType,
          _id: base._id
        }
      };
      mergeContext.mergedDoc.changeElement(changeCmd);
    }
  },
  onDeleteElement: (elementType, elementId, mergeContext) => {
    const deleteCmd = {
      __typename: _HTypes.HDocCommandType.DELETE_ELEMENT,
      path: (0, _HDocument.pathForElementWithId)(mergeContext.mergedDoc, elementType, elementId),
      targetElement: {
        __typename: elementType,
        _id: elementId
      }
    };
    mergeContext.mergedDoc.deleteElement(deleteCmd);
  },
  cmpSiblings: (elementType, base, a, b) => {
    if (a === null && b === null) return 0;
    if (a === null) return 1;
    if (b === null) return -1;
    if (a._id === b._id) return 0;
    if (base && a._id === base._id) return 1;
    if (base && b._id === base._id) return -1;
    return (a._id || 0) < (b._id || 0) ? -1 : (a._id || 0) > (b._id || 0) ? 1 : 0;
  },
  arePositionsCompatible: (elementType, elementId, fromSide, mergeContext) => {
    const leftElement = (0, _HDocument.hasMappedElement)(mergeContext.myDoc().maps, elementType, elementId) ? (0, _HDocument.mappedElement)(mergeContext.myDoc().maps, elementType, elementId) : null;
    const rightElement = (0, _HDocument.hasMappedElement)(mergeContext.theirDoc().maps, elementType, elementId) ? (0, _HDocument.mappedElement)(mergeContext.theirDoc().maps, elementType, elementId) : null;
    if (!(leftElement && rightElement)) return true;

    if (!(0, _HDocument.isNullableId)(leftElement.parentId) && !(0, _HDocument.isNullableId)(rightElement.parentId)) {
      return true;
    }

    if (leftElement.parentId !== rightElement.parentId || leftElement.parentType !== rightElement.parentType) {
      return false;
    }

    return fromSide === _HTypes.ProcessingOrderFrom.both;
  },
  addElement: (elementType, element, parentPath, position, mergeContext) => {
    const insertCmd = {
      __typename: _HTypes.HDocCommandType.INSERT_ELEMENT,
      element: stripChildrenFromElement(mergeContext.mergedDoc.schema, elementType, element),
      parentPath,
      position,
      targetElement: element._id ? {
        __typename: elementType,
        _id: element._id
      } : undefined
    };
    return mergeContext.mergedDoc.insertElement(insertCmd);
  }
};
/**
 * Given a base version of a normalized document, and
 * two later versions of the document, generates a merge
 * tree reporting back if there were conflicts and a diff
 * object to go from myDoc to theirDoc.
 *
 * @param {INormalizedDocument<MapsInterface, U>} baseDoc
 * @param {INormalizedDocument<MapsInterface, U>} myDoc
 * @param {INormalizedDocument<MapsInterface, U>} theirDoc
 * @param {MergeOverridesMap<MapsInterface, U, ElementType>} options
 * @returns {II3MergeResult<MapsInterface, U>}
 */

function threeWayMerge(baseDoc, myDoc, theirDoc, options) {
  const mergedDoc = options && options.onCreateMutableDocument ? options.onCreateMutableDocument(myDoc) : (0, _HDocument.mutableDocument)(myDoc);
  const mergeContext = {
    mergedDoc,
    myElementsMergeState: buildMergeElementsState(baseDoc, myDoc),
    theirElementsMergeState: buildMergeElementsState(baseDoc, theirDoc),
    baseDoc,
    myDoc: createGetterSetter(myDoc),
    theirDoc: createGetterSetter(theirDoc),
    elementsToDelete: [],
    conflicts: {},
    overrides: options && options.elementsOverrides ? options.elementsOverrides : {},
    defaultHooks: defaultHooks
  };

  for (const elementType in baseDoc.maps) {
    // ToDo resolve this type error
    // @ts-expect-error
    mergeContext.conflicts[elementType] = new Map();
  }

  buildMergedTree(mergeContext);
  const updatedDoc = mergedDoc.updatedDocument();
  return {
    mergedDoc: updatedDoc,
    conflicts: mergeContext.conflicts
  };
}

function mergeElementInfo(schema, elementType, baseElement, leftElement, rightElement, excludeFields = []) {
  const mergeResult = {
    mergedElement: _objectSpread({}, baseElement)
  };
  const conflicts = {};
  const cloneElement = mergeResult.mergedElement;

  for (const fieldName in cloneElement) {
    if (fieldName === 'parentId' || fieldName === 'parentType' || fieldName in schema.types[elementType] || excludeFields.indexOf(fieldName) !== -1 || !isDataValue(cloneElement[fieldName])) {
      continue;
    }

    const mergedVal = mergeDataValues(baseElement[fieldName], leftElement[fieldName], rightElement[fieldName]);

    if (isDataValue(mergedVal)) {
      // @ts-expect-error
      cloneElement[fieldName] = mergedVal;
    } else {
      // @ts-expect-error
      cloneElement[fieldName] = mergedVal.mergedValue; // @ts-expect-error

      conflicts[fieldName] = mergedVal;
    }
  }

  return Object.keys(conflicts).length > 0 ? _objectSpread(_objectSpread({}, mergeResult), {}, {
    conflicts
  }) : mergeResult;
}

const getElementTypeUid = (doc, elementType) => `${doc.schema.documentType}.${elementType}`;

const elementTypesOverridesMap = new Map();

function createGetterSetter(value) {
  let val = value;
  return newValue => {
    if (newValue !== undefined) {
      val = newValue;
    }

    return val;
  };
}
/**
 * Creates new IDs for all the elements in the document subTree
 * starting at the given element identified by its type and id.
 *
 * This is a destructive change which is not tracked as usual via
 * the changes array, it is meant to be used only within the merging
 * functions
 *
 * @param {IMutableDocument<MapsInterface, U>} document
 * @param {U} elementType
 * @param {Id} elementId
 * @returns {IMutableDocument<MapsInterface, U>}
 */


function reIdElementSubtree(context, treeToRebase, elementType, elementId) {
  const document = treeToRebase === _HTypes.ProcessingOrderFrom.left ? context.myDoc() : context.theirDoc();
  const newIds = new Map();
  const elUid = getElementUid(elementType, elementId);
  const changedDocument = (0, _HDocument.cloneNormalizedDocument)(document); // First I generate Ids for all the elements in the subtree

  (0, _HVisit.visitDocument)(document, (doc, nodeType, nodeId) => {
    const nodeUid = getElementUid(nodeType, nodeId);
    const newId = (0, _uuid.v1)();

    if (!newIds.has(nodeUid)) {
      newIds.set(nodeUid, newId);

      if (treeToRebase === _HTypes.ProcessingOrderFrom.left) {
        context.myElementsMergeState.set(getElementUid(nodeType, newId), context.myElementsMergeState.get(getElementUid(nodeType, nodeId)));
      } else {
        context.theirElementsMergeState.set(getElementUid(nodeType, newId), context.theirElementsMergeState.get(getElementUid(nodeType, nodeId)));
      }
    }
  }, {}, _HTypes.DocumentVisitTraversal.DEPTH_FIRST, elementType, elementId);
  const rebasingRootElement = (0, _HDocument.mappedElement)(changedDocument.maps, elementType, elementId);
  const rebasedRootId = newIds.get(elUid);

  if (rebasingRootElement.parentId && rebasingRootElement.parentType) {
    const parent = (0, _HDocument.mappedElement)(changedDocument.maps, rebasingRootElement.parentType, rebasingRootElement.parentId);
    const parentToChildField = (0, _HDocument.parentToChildFieldName)(changedDocument, rebasingRootElement.parentType, elementType);

    const newParent = _objectSpread(_objectSpread({}, parent), {}, {
      [parentToChildField]: parent[parentToChildField].slice()
    });

    const oldIndex = newParent[parentToChildField].indexOf(elementId);

    if (oldIndex !== -1) {
      newParent[parentToChildField][oldIndex] = rebasedRootId;
    }

    changedDocument.maps[rebasingRootElement.parentType].set(rebasingRootElement.parentId, newParent);
  } else {
    changedDocument.rootId = rebasedRootId;
  } // Second I go depth first down the elements and change the ids
  // of each one of them


  (0, _HVisit.visitDocument)(changedDocument, (doc, elementType, elementId) => {
    const element = (0, _HDocument.mappedElement)(doc.maps, elementType, elementId);
    const reIdUid = getElementUid(elementType, elementId);

    const reIdedElement = _objectSpread(_objectSpread({}, element), {}, {
      _id: newIds.get(reIdUid)
    });

    const nodeSchema = doc.schema.types[elementType];
    const reIdedParentId = element.parentId !== null && newIds.has(getElementUid(element.parentType, element.parentId)) ? newIds.get(getElementUid(element.parentType, element.parentId)) : null;

    for (const linkField in nodeSchema) {
      if (linkField === 'parentId' && reIdedParentId !== null) {
        reIdedElement.parentId = reIdedParentId;
        continue;
      }

      const linkFieldProps = nodeSchema[linkField];

      if (Array.isArray(linkFieldProps)) {
        const {
          __schemaType
        } = linkFieldProps[0];
        reIdedElement[linkField] = reIdedElement[linkField].map(existingId => newIds.get(getElementUid(__schemaType, existingId)));
      } else {
        const {
          __schemaType
        } = linkFieldProps;
        reIdedElement[linkField] = newIds.get(getElementUid(__schemaType, reIdedElement[linkField]));
      }
    }

    doc.maps[elementType].set(reIdedElement._id, reIdedElement);
    doc.maps[elementType].delete(elementId);
  }, {}, _HTypes.DocumentVisitTraversal.DEPTH_FIRST, elementType, elementId);

  if (treeToRebase === _HTypes.ProcessingOrderFrom.left) {
    context.myDoc(changedDocument);
  } else {
    context.theirDoc(changedDocument);
  }

  return {
    doc: changedDocument,
    newElementId: newIds.get(elUid)
  };
}
/**
 * Provides the functions to use for an elementType
 * @param {II3WMergeContext<MapsInterface, U>} context
 * @param {U} elementType
 * @returns {IMergeElementOverrides<MapsInterface, U, MapsInterface[typeof elementType]>}
 */


function fnsForElementType(context, elementType) {
  const typeUid = getElementTypeUid(context.myDoc(), elementType);
  let overridableFunctions = elementTypesOverridesMap.get(typeUid);

  if (!overridableFunctions) {
    const {
      overrides
    } = context;
    const elementOverrides = overrides && elementType in overrides ? overrides[elementType] : {};
    overridableFunctions = {
      onIncompatibleElementVersions: elementOverrides.onIncompatibleElementVersions ? elementOverrides.onIncompatibleElementVersions : (elementId, parentPath, position, versionMoved, mergeContext) => context.defaultHooks.onIncompatibleElementVersions(elementType, elementId, parentPath, position, versionMoved, mergeContext),
      moveToMergePosition: elementOverrides.moveToMergePosition ? elementOverrides.moveToMergePosition : (elementId, toParentPath, toPosition, mergeContext) => context.defaultHooks.moveToMergePosition(elementType, elementId, toParentPath, toPosition, mergeContext),
      mergeElementInfo: elementOverrides.mergeElementInfo ? elementOverrides.mergeElementInfo : (base, a, b, mergeContext, ignoreFields) => context.defaultHooks.mergeElementInfo(mergeContext, elementType, base, a, b, ignoreFields),
      onDeleteElement: elementOverrides.onDeleteElement ? elementOverrides.onDeleteElement : (elementId, mergeContext) => context.defaultHooks.onDeleteElement(elementType, elementId, mergeContext),
      cmpSiblings: elementOverrides.cmpSiblings ? elementOverrides.cmpSiblings : (base, a, b, mergeCtx) => context.defaultHooks.cmpSiblings(elementType, base, a, b, mergeCtx),
      arePositionsCompatible: elementOverrides.arePositionsCompatible ? elementOverrides.arePositionsCompatible : (elementId, fromSide, mergeContext) => context.defaultHooks.arePositionsCompatible(elementType, elementId, fromSide, mergeContext),
      addElement: elementOverrides.addElement ? elementOverrides.addElement : (element, parentPath, position, mergeContext) => context.defaultHooks.addElement(elementType, element, parentPath, position, mergeContext)
    };
    elementTypesOverridesMap.set(typeUid, overridableFunctions);
  }

  return overridableFunctions;
}
/**
 * Creates the desired shape of the merged three during a three-way
 * merge of a NormalizedDocument.
 *
 * @param {II3WMergeContext<MapsInterface, U>} mergeCtx
 * @returns {IMutableDocument<MapsInterface, U>}
 */


function buildMergedTree(mergeCtx) {
  const {
    myDoc: left,
    theirDoc: right,
    mergedDoc,
    baseDoc
  } = mergeCtx;

  for (const nodeQueue = [[mergedDoc.rootType, mergedDoc.rootId]], nodesInQueue = new Set([getElementUid(mergedDoc.rootType, mergedDoc.rootId)]); nodeQueue.length > 0;) {
    const [nodeType, nodeId] = nodeQueue.shift();
    const {
      mergeElementInfo
    } = fnsForElementType(mergeCtx, nodeType);
    const baseEl = (0, _HDocument.hasMappedElement)(baseDoc.maps, nodeType, nodeId) ? (0, _HDocument.mappedElement)(baseDoc.maps, nodeType, nodeId) : null;
    const leftEl = (0, _HDocument.hasMappedElement)(left().maps, nodeType, nodeId) ? (0, _HDocument.mappedElement)(left().maps, nodeType, nodeId) : null;
    const rightEl = (0, _HDocument.hasMappedElement)(right().maps, nodeType, nodeId) ? (0, _HDocument.mappedElement)(right().maps, nodeType, nodeId) : null;
    mergeElementInfo(baseEl, leftEl, rightEl, mergeCtx);
    const nodeSchema = mergedDoc.schema.types[nodeType];

    for (const linkField in nodeSchema) {
      if (linkField === 'parentId') continue;
      const linkFieldProps = nodeSchema[linkField];

      if (Array.isArray(linkFieldProps)) {
        const elementsToPush = mergeLinkedArray(mergeCtx, nodeType, nodeId, linkFieldProps[0].__schemaType, linkField).filter(([nodeType, nodeId]) => {
          const nodeUid = getElementUid(nodeType, nodeId);

          if (nodesInQueue.has(nodeUid)) {
            return false;
          }

          nodesInQueue.add(nodeUid);
          return true;
        });
        nodeQueue.push(...elementsToPush);
      } // We only expect children arrays, we don't deal with one to one
      // linked fields. They should be modelled as single element arrays,
      // so that in merges we could clone copies if needed.

    }
  }

  const elementsToDelete = [];
  (0, _HVisit.visitDocument)(mergedDoc, (doc, nodeType, nodeId) => {
    const nodeUid = getElementUid(nodeType, nodeId);
    const existsLeft = (0, _HDocument.hasMappedElement)(mergeCtx.myDoc().maps, nodeType, nodeId);
    const editedLeft = existsLeft ? mergeCtx.myElementsMergeState.get(nodeUid).isInEditedPath : false;
    const existsRight = (0, _HDocument.hasMappedElement)(mergeCtx.theirDoc().maps, nodeType, nodeId);
    const editedRight = existsRight ? mergeCtx.theirElementsMergeState.get(nodeUid).isInEditedPath : false;

    if ((!existsLeft || !existsRight) && !editedLeft && !editedRight) {
      elementsToDelete.push([nodeType, nodeId]);
    }
  }, {}, _HTypes.DocumentVisitTraversal.DEPTH_FIRST);

  for (const [nodeType, nodeId] of elementsToDelete) {
    const {
      onDeleteElement
    } = fnsForElementType(mergeCtx, nodeType);
    onDeleteElement(nodeId, mergeCtx);
  }

  return mergedDoc;
}

function isOkMergeZone(obj) {
  return obj && obj.ok && Array.isArray(obj.ok);
}

function isConflictMergeZone(obj) {
  return obj && typeof obj.conflict === 'object' && Array.isArray(obj.conflict.a) && Array.isArray(obj.conflict.b) && Array.isArray(obj.conflict.o) && typeof obj.conflict.aIndex === 'number' && typeof obj.conflict.bIndex === 'number' && typeof obj.conflict.oIndex === 'number';
}
/**
 * Creates an iterator over a three-way merging array
 * of children IDs.
 *
 * The iterator allows changing an id to a new one if this was
 * needed during the merging and we want to ensure the new rebased
 * value is used straight away.
 */


_Symbol$iterator = Symbol.iterator;

class NextSiblingToProcessIterator {
  constructor(mergeCtx, parentType, parentId, childType, linkedArrayField) {
    _defineProperty(this, "baseArray", void 0);

    _defineProperty(this, "_mergingArray", void 0);

    _defineProperty(this, "_leftArray", void 0);

    _defineProperty(this, "_rightArray", void 0);

    _defineProperty(this, "mergeZones", void 0);

    _defineProperty(this, "mergeCtx", void 0);

    _defineProperty(this, "childType", void 0);

    _defineProperty(this, "parentType", void 0);

    _defineProperty(this, "next", () => {
      if (this.mergeZones.length === 0) {
        return {
          done: true,
          value: undefined
        };
      }

      const mergeZone = this.mergeZones[0];

      if (isOkMergeZone(mergeZone)) {
        const nextId = mergeZone.ok.shift();
        const isInLeft = this._leftArray.indexOf(nextId) !== -1;
        const isInRight = this._rightArray.indexOf(nextId) !== -1;
        const nextFrom = isInLeft && isInRight ? _HTypes.ProcessingOrderFrom.both : isInLeft ? _HTypes.ProcessingOrderFrom.left : _HTypes.ProcessingOrderFrom.right;
        const nextValue = {
          _id: nextId,
          from: nextFrom
        };

        if (mergeZone.ok.length === 0) {
          this.mergeZones.shift();
        }

        return {
          done: false,
          value: nextValue
        };
      } else if (isConflictMergeZone(mergeZone)) {
        let nextValue = null;

        if (mergeZone.conflict.a.length === 0 && mergeZone.conflict.b.length > 0) {
          nextValue = {
            _id: mergeZone.conflict.b.shift(),
            from: _HTypes.ProcessingOrderFrom.right
          };
        } else if (mergeZone.conflict.b.length === 0 && mergeZone.conflict.a.length > 0) {
          nextValue = {
            _id: mergeZone.conflict.a.shift(),
            from: _HTypes.ProcessingOrderFrom.left
          };
        } else {
          const {
            conflict: {
              a,
              b,
              o
            }
          } = mergeZone;
          const leftId = a[0];
          const rightId = b[0];

          if (leftId === rightId) {
            nextValue = {
              _id: mergeZone.conflict.a.shift(),
              from: _HTypes.ProcessingOrderFrom.left
            };
            mergeZone.conflict.b.shift();
          } else {
            const baseEl = o.length > 0 ? (0, _HDocument.mappedElement)(this.mergeCtx.baseDoc.maps, this.childType, o[0]) : null;
            const leftEl = leftId ? (0, _HDocument.mappedElement)(this.mergeCtx.myDoc().maps, this.childType, leftId) : null;
            const rightEl = rightId ? (0, _HDocument.mappedElement)(this.mergeCtx.theirDoc().maps, this.childType, rightId) : null;
            const {
              cmpSiblings
            } = fnsForElementType(this.mergeCtx, this.parentType);
            const siblingsComparison = cmpSiblings(baseEl, leftEl, rightEl, this.mergeCtx);
            nextValue = {
              _id: siblingsComparison === 0 || siblingsComparison < 1 ? a[0] : b[0],
              from: siblingsComparison === 0 ? _HTypes.ProcessingOrderFrom.both : siblingsComparison < 1 ? _HTypes.ProcessingOrderFrom.left : _HTypes.ProcessingOrderFrom.right
            };

            if (siblingsComparison <= 0) {
              a.shift();
            }

            if (siblingsComparison >= 0) {
              b.shift();
            }

            if (o.length > 0) {
              o.shift();
            }
          }
        }

        if (mergeZone.conflict.a.length === 0 && mergeZone.conflict.b.length === 0) {
          this.mergeZones.shift();
        }

        return nextValue ? {
          done: false,
          value: nextValue
        } : {
          done: true,
          value: undefined
        };
      } else {
        return {
          done: true,
          value: undefined
        };
      }
    });

    this.mergeCtx = mergeCtx;
    this.childType = childType;
    this.parentType = parentType;
    this.baseArray = (0, _HDocument.hasMappedElement)(mergeCtx.baseDoc.maps, parentType, parentId) ? (0, _HDocument.mappedElement)(mergeCtx.baseDoc.maps, parentType, parentId)[linkedArrayField] : [];
    this._mergingArray = (0, _HDocument.mappedElement)(mergeCtx.mergedDoc.maps, parentType, parentId)[linkedArrayField].slice();
    this._leftArray = (0, _HDocument.hasMappedElement)(mergeCtx.myDoc().maps, parentType, parentId) ? (0, _HDocument.mappedElement)(mergeCtx.myDoc().maps, parentType, parentId)[linkedArrayField].slice() : [];
    this._rightArray = (0, _HDocument.hasMappedElement)(mergeCtx.theirDoc().maps, parentType, parentId) ? (0, _HDocument.mappedElement)(mergeCtx.theirDoc().maps, parentType, parentId)[linkedArrayField].slice() : [];
    this.mergeZones = (0, _nodeDiff.diff3Merge)(this._leftArray, this.baseArray, this._rightArray);
  }

  reId(existingId, newId, sideToReId) {
    if (sideToReId === _HTypes.ProcessingOrderFrom.left || sideToReId === _HTypes.ProcessingOrderFrom.both) {
      for (let i = 0; i < this._leftArray.length; i++) {
        if (this._leftArray[i] === existingId) {
          this._leftArray[i] = newId;
        }
      }
    }

    if (sideToReId === _HTypes.ProcessingOrderFrom.right || sideToReId === _HTypes.ProcessingOrderFrom.both) {
      for (let i = 0; i < this._rightArray.length; i++) {
        if (this._rightArray[i] === existingId) {
          this._rightArray[i] = newId;
        }
      }
    }

    for (const mergeZone of this.mergeZones) {
      if (isOkMergeZone(mergeZone)) {
        for (let i = 0; i < mergeZone.ok.length; i++) {
          if (mergeZone.ok[i] === existingId) {
            mergeZone.ok[i] = newId;
          }
        }
      } else if (isConflictMergeZone(mergeZone)) {
        for (let i = 0; i < mergeZone.conflict.a.length; i++) {
          if (mergeZone.conflict.a[i] === existingId) {
            mergeZone.conflict.a[i] = newId;
          }
        }

        for (let i = 0; i < mergeZone.conflict.b.length; i++) {
          if (mergeZone.conflict.b[i] === existingId) {
            mergeZone.conflict.b[i] = newId;
          }
        }

        for (let i = 0; i < mergeZone.conflict.o.length; i++) {
          if (mergeZone.conflict.o[i] === existingId) {
            mergeZone.conflict.o[i] = newId;
          }
        }
      }
    }
  }

  [_Symbol$iterator]() {
    return this;
  }

}
/**
 * Given a parent node and the name a linked field representing an array
 * of ids to elements of type childType, creates a merged version
 * of the array in mergeCtx.mergedDoc
 *
 * @param {II3WMergeContext<MapsInterface, U>} mergeCtx
 * @param {U} parentType
 * @param {Id} parentId
 * @param {U} childType
 * @param {AllMappedTypesFields<MapsInterface>} linkedArrayField
 * @returns {Array<[U, Id]>}
 */


function mergeLinkedArray(mergeCtx, parentType, parentId, childType, linkedArrayField) {
  const {
    mergedDoc,
    myDoc,
    theirDoc,
    myElementsMergeState,
    theirElementsMergeState
  } = mergeCtx;
  const {
    addElement,
    moveToMergePosition,
    arePositionsCompatible,
    onIncompatibleElementVersions
  } = fnsForElementType(mergeCtx, childType);
  const idsToProcessIterator = new NextSiblingToProcessIterator(mergeCtx, parentType, parentId, childType, linkedArrayField);
  const childrenToQueue = [];
  const parentPath = (0, _HDocument.pathForElementWithId)(mergedDoc, parentType, parentId);

  for (let i = 0, il = 0, ir = 0, mergedIndex = 0, nextId = idsToProcessIterator.next(); !nextId.done; nextId = idsToProcessIterator.next(), i++) {
    const mergedIndexAtStart = mergedIndex;
    const baseArray = (0, _HDocument.mappedElement)(mergedDoc.maps, parentType, parentId)[linkedArrayField];
    const leftArray = (0, _HDocument.hasMappedElement)(mergeCtx.myDoc().maps, parentType, parentId) ? (0, _HDocument.mappedElement)(mergeCtx.myDoc().maps, parentType, parentId)[linkedArrayField] : [];
    const rightArray = (0, _HDocument.hasMappedElement)(mergeCtx.theirDoc().maps, parentType, parentId) ? (0, _HDocument.mappedElement)(mergeCtx.theirDoc().maps, parentType, parentId)[linkedArrayField] : [];
    const baseChildId = baseArray.length > mergedIndex ? baseArray[mergedIndex] : null;
    const leftChildId = leftArray.length > il ? leftArray[il] : null;
    const rightChildId = rightArray.length > ir ? rightArray[ir] : null;
    const {
      _id: childId,
      from: childFrom
    } = nextId.value;
    if (childId === leftChildId) il++;
    if (childId === rightChildId) ir++;
    const baseElement = (0, _HDocument.hasMappedElement)(mergedDoc.maps, childType, childId) ? (0, _HDocument.mappedElement)(mergedDoc.maps, childType, childId) : null;
    const leftElement = (0, _HDocument.hasMappedElement)(myDoc().maps, childType, childId) ? (0, _HDocument.mappedElement)(myDoc().maps, childType, childId) : null;
    const rightElement = (0, _HDocument.hasMappedElement)(theirDoc().maps, childType, childId) ? (0, _HDocument.mappedElement)(theirDoc().maps, childType, childId) : null;
    const childUid = getElementUid(childType, childId);

    if (baseElement) {
      const leftState = myElementsMergeState.get(childUid);
      const rightState = theirElementsMergeState.get(childUid);

      if (leftState && leftState.isInEditedPath && rightState && rightState.isInEditedPath) {
        if (arePositionsCompatible(childId, childFrom, mergeCtx)) {
          if (!leftState.hasPositionBeenProcessed) {
            if (baseChildId !== childId) {
              moveToMergePosition(childId, parentPath, [linkedArrayField, mergedIndex], mergeCtx);
            }

            mergedIndex++;
            leftState.hasPositionBeenProcessed = true;
            rightState.hasPositionBeenProcessed = true;
          }
        } else {
          const {
            advancedMergingIndex,
            rebasedIds
          } = onIncompatibleElementVersions(childId, parentPath, [linkedArrayField, mergedIndex], childFrom === _HTypes.ProcessingOrderFrom.right ? _HTypes.ProcessingOrderFrom.right : _HTypes.ProcessingOrderFrom.left, mergeCtx);

          for (const {
            _id: oldId,
            newId,
            rebasedSide
          } of rebasedIds) {
            idsToProcessIterator.reId(oldId, newId, rebasedSide);
          }

          if (advancedMergingIndex) mergedIndex++;
        }
      } else if (leftState && leftState.isInEditedPath || rightState && rightState.isInEditedPath) {
        // One of the two versions of the node is in an edited path, we
        // move it to the current position if the edited node is the one we
        // are meeting now
        if (leftState && leftState.isInEditedPath && (childFrom === _HTypes.ProcessingOrderFrom.left || childFrom === _HTypes.ProcessingOrderFrom.both)) {
          if (baseChildId !== childId) {
            moveToMergePosition(childId, parentPath, [linkedArrayField, mergedIndex], mergeCtx);
            leftState.hasPositionBeenProcessed = true;
            if (rightState) rightState.hasPositionBeenProcessed = true;
          }

          mergedIndex++;
        } else if (rightState && rightState.isInEditedPath && (childFrom === _HTypes.ProcessingOrderFrom.right || childFrom === _HTypes.ProcessingOrderFrom.both)) {
          if (baseChildId !== childId) {
            moveToMergePosition(childId, parentPath, [linkedArrayField, mergedIndex], mergeCtx);
          }

          rightState.hasPositionBeenProcessed = true;
          if (leftState) leftState.hasPositionBeenProcessed = true;
          mergedIndex++;
        }
      } else if (!leftElement || !rightElement) {
        // The node is not present in at least one of the nodes, so
        // it should be deleted. Right now we don't do anything, it
        // will be removed at the end of the merge process. We only marked
        // the node as processed.
        if (leftState) {
          leftState.hasPositionBeenProcessed = true;
          leftState.haveInfoAndChildrenBeenProcessed = true;
        }

        if (rightState) {
          rightState.hasPositionBeenProcessed = true;
          rightState.haveInfoAndChildrenBeenProcessed = true;
        }
      } else {
        // The node is present in both nodes but it hasn't been edited.
        // We add it on the first occurrence
        if (!leftState.hasPositionBeenProcessed) {
          if (childId !== baseChildId) {
            moveToMergePosition(childId, parentPath, [linkedArrayField, mergedIndex], mergeCtx);
          }

          mergedIndex++;
          leftState.hasPositionBeenProcessed = true;
          rightState.hasPositionBeenProcessed = true;
          leftState.haveInfoAndChildrenBeenProcessed = true;
          rightState.haveInfoAndChildrenBeenProcessed = true;
        }
      }
    } else {
      const elementToAdd = leftElement || rightElement;
      addElement(elementToAdd, parentPath, [linkedArrayField, mergedIndex], mergeCtx);
      const elementState = myElementsMergeState.get(childUid) || theirElementsMergeState.get(childUid);
      elementState.hasPositionBeenProcessed = true;
      mergedIndex++;
    }

    if (mergedIndex > mergedIndexAtStart) {
      childrenToQueue.push([childType, childId]);
    }
  }

  return childrenToQueue;
}

function getElementUid(elementType, elemendId) {
  return `${elementType}:${elemendId}`;
}
/**
 * Creates a map of IMergeElementsState records for each element in the
 * document, in preparation for a merge.
 *
 * @param {INormalizedDocument<MapsInterface, U>} baseDoc
 * @param {INormalizedDocument<MapsInterface, U>} laterDoc
 * @returns {Map<string, IMergeElementsState>}
 */


function buildMergeElementsState(baseDoc, laterDoc) {
  const mergeElementsState = new Map();
  (0, _HVisit.visitDocument)(laterDoc, (doc, nodeType, nodeId) => {
    mergeElementsState.set(getElementUid(nodeType, nodeId), {
      hasPositionBeenProcessed: false,
      haveInfoAndChildrenBeenProcessed: false,
      isInBaseTree: (0, _HDocument.hasMappedElement)(baseDoc.maps, nodeType, nodeId),
      isInEditedPath: false,
      mergedElementId: nodeId
    });
  }, {});
  const branchChanges = (0, _HDiff.diff)(baseDoc, laterDoc);

  for (const branchChange of branchChanges) {
    if (branchChange.__typename === _HTypes.HDocCommandType.DELETE_ELEMENT) {
      // The element is not in later tree nor in the mergeElementsState, I can
      // just skip it
      continue;
    }

    const {
      __typename,
      _id
    } = branchChange.targetElement;

    for (let nextType = __typename, nextId = _id, i = 0; nextId !== null && nextType !== null && i < 10000; i++) {
      const element = (0, _HDocument.mappedElement)(laterDoc.maps, nextType, nextId);

      if (!(0, _HDocument.isParentedId)(element)) {
        nextId = null;
      } else {
        const elementUid = getElementUid(nextType, nextId);
        const elementState = mergeElementsState.get(elementUid);

        if (elementState) {
          mergeElementsState.set(elementUid, _objectSpread(_objectSpread({}, elementState), {}, {
            isInEditedPath: true
          }));
        }

        nextId = element.parentId;
        nextType = element.parentType;
      }
    }
  }

  return mergeElementsState;
}
/**
 * Creates a shallow copy of the element record with the given type and id, and strips it of
 * all its children.
 *
 * @param {INormalizedDocument<MapsInterface, U> | IMutableDocument<MapsInterface, U>} doc
 * @param {U} elementType
 * @param {Id} elementId
 * @returns {T}
 */


function stripChildrenFromElement(schema, elementType, element) {
  const clonedElement = _objectSpread({}, element);

  const fieldsMap = schema.types[elementType];

  for (const fieldName in fieldsMap) {
    if (fieldName === 'parentId') continue;

    if (Array.isArray(fieldsMap[fieldName])) {
      clonedElement[fieldName] = [];
    } else {
      clonedElement[fieldName] = null;
    }
  }

  return clonedElement;
}
//# sourceMappingURL=HMerge3.js.map