"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.visitDocument = visitDocument;

var _HTypes = require("./HTypes");

var _HDocument = require("./HDocument");

/**
 * Traversal of a normalized document, calling
 * for each node the visitor function passed as parameter.
 *
 * The traversal is breadth first by default, unless the
 * goBreadthFirst parameter is false in which case its depthFirst
 *
 * @param {IDocumentSchema<MapsInterface, V>} docSchema
 * @param {INormalizedDocument<MapsInterface, U, V>} doc
 * @param {IVisitor<MapsInterface, U, V>} onNodeVisit
 * @param {Context} context
 */
function visitDocument(doc, onNodeVisit, context, traversal = _HTypes.DocumentVisitTraversal.BREADTH_FIRST, elementType, elementId) {
  for (const nodesToVisit = traversal === _HTypes.DocumentVisitTraversal.BREADTH_FIRST ? breadthFirstNodes(doc, elementType || doc.rootType, elementId || doc.rootId) : depthFirstNodes(doc, elementType || doc.rootType, elementId || doc.rootId); nodesToVisit.length > 0;) {
    const [nextType, nextId] = nodesToVisit.shift();
    onNodeVisit(doc, nextType, nextId, context || {});
  }
}

function breadthFirstNodes(doc, nodeType, nodeId, nodesVisited = new Set()) {
  const nodeUid = `${nodeType}:${nodeId}`;

  if (nodesVisited.has(nodeUid)) {
    return [];
  } else {
    nodesVisited.add(nodeUid);
  }

  const element = (0, _HDocument.mappedElement)(doc.maps, nodeType, nodeId);
  if (!(0, _HDocument.isParentedId)(element)) return [];
  const nodeList = [[nodeType, nodeId]];
  const nodeSchema = doc.schema.types[nodeType];

  for (const linkField in nodeSchema) {
    if (linkField === 'parentId') continue;
    const linkFieldProps = nodeSchema[linkField];

    if (Array.isArray(linkFieldProps)) {
      const {
        __schemaType
      } = linkFieldProps[0];
      const fieldIds = element[linkField];

      for (const fieldId of fieldIds) {
        nodeList.push([__schemaType, fieldId]);
      }
    } else {
      const {
        __schemaType
      } = linkFieldProps;
      nodeList.push([__schemaType, element[linkField]]);
    }
  }

  const lengthWithChildren = nodeList.length;

  for (let i = 1; i < lengthWithChildren; i++) {
    nodeList.push(...breadthFirstNodes(doc, nodeList[i][0], nodeList[i][1], nodesVisited));
  }

  return nodeList;
}

function depthFirstNodes(doc, nodeType, nodeId, nodesVisited = new Set()) {
  const nodeUid = `${nodeType}:${nodeId}`;

  if (nodesVisited.has(nodeUid)) {
    return [];
  } else {
    nodesVisited.add(nodeUid);
  }

  const nodeList = [[nodeType, nodeId]];
  const element = (0, _HDocument.mappedElement)(doc.maps, nodeType, nodeId);
  if (!(0, _HDocument.isParentedId)(element)) return [];
  const nodeSchema = doc.schema.types[nodeType];

  for (const linkField in nodeSchema) {
    if (linkField === 'parentId') continue;
    const linkFieldProps = nodeSchema[linkField];

    if (Array.isArray(linkFieldProps)) {
      const {
        __schemaType
      } = linkFieldProps[0];
      const fieldIds = element[linkField];

      for (const fieldId of fieldIds) {
        nodeList.unshift(...depthFirstNodes(doc, __schemaType, fieldId, nodesVisited));
      }
    } else {
      const {
        __schemaType
      } = linkFieldProps;
      nodeList.unshift(...depthFirstNodes(doc, __schemaType, element[linkField], nodesVisited));
    }
  }

  return nodeList;
}
//# sourceMappingURL=HVisit.js.map